/***************************************************************************
                          utils.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Apr 25 1999
    copyright            : (C) 2000 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <math.h>
#include <dlfcn.h>
#include <qlineedit.h>
#include <qmultilinedit.h>
#include <qfileinfo.h>
#include <qdir.h>
#include <qtextstream.h>
#include <klocale.h>
#include <kapp.h>
#include <kurl.h>
#include <kfiledialog.h>

#if (KDE_VERSION_MAJOR > 1) || (KDE_VERSION_MINOR > 1)
  #include <kmessagebox.h>
  #include <kio_job.h>
  #include <kio_netaccess.h>
  #include <kglobal.h>
#else
  #include <kmsgbox.h>
  #include <kfm.h>
#endif

#include "kpldoc.h"
#include "funcdlg.h"

bool getVal(QWidget *parent, QLineEdit *edit, double *val,
            const char *errmsg)
{
  bool ok = true;
  QString s = edit->text();
  *val = s.toDouble(&ok);
  if (!ok)
#if (KDE_VERSION_MAJOR > 1) || (KDE_VERSION_MINOR > 1)
    KMessageBox::sorry(parent, errmsg);
#else
    KMsgBox::message(parent, i18n("Sorry"), errmsg, KMsgBox::EXCLAMATION);
#endif
  return ok;
}

void getFile(QLineEdit *e, KplDoc *m)
{
  QString s = KFileDialog::getOpenFileName(m->currentDir(), "*.so");
  if (!s.isEmpty()) {
    e->setText(s);
    QFileInfo fi(s);
    m->setCurrentDir(fi.dirPath(true));
  }
}

void getFunc(QWidget *parent, QLineEdit *name, QLineEdit *func)
{
  FILE *f;

  QFileInfo fi(name->text());
  QString path = fi.dirPath(true) + "/" + fi.baseName() + ".def";
  if ((f = fopen(path, "r"))) {
    ChooseFuncDlg dlg(parent, f, func);
    dlg.exec();
  } else
#if (KDE_VERSION_MAJOR > 1) || (KDE_VERSION_MINOR > 1)
    KMessageBox::error(parent, i18n("while trying to open module definition file"));
#else
    KMsgBox::message(parent, i18n("Error"),
                     i18n("while trying to open module definition file"),
                     KMsgBox::EXCLAMATION);
#endif
}

void getPar(QWidget *parent, double *p, KplDoc *m)
{
  int i;
  char s[81];
  FILE *f;

  QString url = KFileDialog::getOpenFileURL(m->currentDir(), "*.par");
  if (!url.isEmpty()) {
    QString fn;
#if (KDE_VERSION_MAJOR > 1) || (KDE_VERSION_MINOR > 1)
    if (KIONetAccess::download(url, fn)) {
#else
    if (KFM::download(url, fn)) {
#endif
      if ((f = fopen(fn, "r"))) {
        for (i = 0; i < NPMAX; i++)
          p[i] = 0.0;
        i = -1;
        while (fgets(s, 80, f) && (++i < NPMAX))
          if (sscanf(s, "%lg", &p[i]) != 1)
            break;
        fclose(f);
        KURL u = KURL(url);
        if (u.isLocalFile()) {
          QFileInfo fi(fn);
          m->setCurrentDir(fi.dirPath(true));
        } else
#if (KDE_VERSION_MAJOR > 1) || (KDE_VERSION_MINOR > 1)
          m->setCurrentDir(u.protocol() + ":" + u.host() + u.directory());
      } else
        KMessageBox::error(parent, i18n("while trying to open parameter file"));
      KIONetAccess::removeTempFile(fn);
#else
          m->setCurrentDir(u.directoryURL());
      } else
        KMsgBox::message(parent, i18n("Error"),
                         i18n("while trying to open parameter file"),
                         KMsgBox::EXCLAMATION);
      KFM::removeTempFile(fn);
#endif
    }
  }
}

void savePar(QWidget *parent, double *p, KplDoc *m)
{
  FILE *f;

  QString url = KFileDialog::getSaveFileURL(m->currentDir(), "*.par");
  if (!url.isEmpty()) {
    KURL u = KURL(url);
    if (u.isLocalFile()) {
      url = u.path();
      QFileInfo fi(url);
      if (fi.exists())
#if (KDE_VERSION_MAJOR > 1) || (KDE_VERSION_MINOR > 1)
        if (KMessageBox::warningYesNo(parent, url + "\n" +
                                       i18n("A file with this name exists already.\n"
                                            "Do you want to overwrite it?")) != KMessageBox::Yes)
#else
        if (KMsgBox::yesNo(parent, i18n("Warning"), url + "\n" +
                         i18n("A file with this name exists already.\n"
                              "Do you want to overwrite it?"),
                         KMsgBox::EXCLAMATION) != 1)
#endif
          return;
      f = fopen(url, "w");
    } else
      f = fopen(QDir::homeDirPath() + "/kpltmp.tmp", "w");
    if (f) {
      QString s;
      for (int i = 0; i < NPMAX; i++) {
        s.setNum(p[i], m->aut.format, m->aut.prec);
        fprintf(f, "%s\n", s.data());
      }
      fclose(f);
      if (u.isLocalFile()) {
        QFileInfo fi(url);
        m->setCurrentDir(fi.dirPath(true));
      } else {
#if (KDE_VERSION_MAJOR > 1) || (KDE_VERSION_MINOR > 1)
        m->setCurrentDir(u.protocol() + ":" + u.host() + u.directory());
        m->kIO = new KIOJob;
#else
        m->setCurrentDir(u.directoryURL());
        m->kIO = new KFM;
#endif
        m->connectIOfinished();
        m->kIO->copy(QDir::homeDirPath() + "/kpltmp.tmp", url);
      }
    } else
#if (KDE_VERSION_MAJOR > 1) || (KDE_VERSION_MINOR > 1)
      KMessageBox::error(parent, i18n("while trying to open parameter file"));
#else
      KMsgBox::message(parent, i18n("Error"),
                       i18n("while trying to open parameter file"),
                       KMsgBox::EXCLAMATION);
#endif
  }
}

void showSource(QMultiLineEdit *ml, QFile *f, const char *func)
{
  int i = 0, line = 0, col = 0;
  QString s;

  ml->setAutoUpdate(false);
#if (KDE_VERSION_MAJOR > 1) || (KDE_VERSION_MINOR > 1)
  ml->setFont(KGlobal::fixedFont());
#else
  ml->setFont(kapp->fixedFont);
#endif
  QTextStream t(f);
  while (!t.eof()) {
    s = t.readLine();
    if (func && !line) {
      col = s.find(QString(" ") + func + "(double");
      if (col != -1)
        line = i;
    }
    ml->append(s);
    i++;
  }
  f->close();
  ml->setAutoUpdate(true);
  ml->repaint();
  if (line) {
    ml->setCursorPosition(line + 15, col);
    ml->setCursorPosition(line, ++col);
    ml->setCursorPosition(line, col + strlen(func), true);
  }
  ml->setReadOnly(true);
  ml->setFocus();
}

bool getFuncAddr(QString path, QString name, void **hmod,
                 double (**fkt)(double, double *))
{
  char *error;
  QString msg;

  *hmod = 0;
  *fkt = 0;
  if (!path.isEmpty()) {
    if (!(*hmod = dlopen(path, RTLD_NOW))) {
      msg.sprintf(i18n("while trying to open\nfile %s:\n%s"),
                  path.data(), dlerror());
#if (KDE_VERSION_MAJOR > 1) || (KDE_VERSION_MINOR > 1)
      KMessageBox::error(0, msg);
#else
      KMsgBox::message(0, i18n("Error"), msg, KMsgBox::EXCLAMATION);
#endif
      return true;
    } else
      if (!name.isEmpty())
        *fkt = (double (*)(double, double *)) dlsym(*hmod, name);
        if ((error = (char *) dlerror()) != 0) {
          msg.sprintf(i18n("while trying to determine the "
                           "address of\nfunction %s:\n%s"),
                      name.data(), error);
#if (KDE_VERSION_MAJOR > 1) || (KDE_VERSION_MINOR > 1)
          KMessageBox::error(0, msg);
#else
          KMsgBox::message(0, i18n("Error"), msg, KMsgBox::EXCLAMATION);
#endif
          dlclose(*hmod);
          *hmod = 0;
          return true;
        } else
          return false;
  }
#if (KDE_VERSION_MAJOR > 1) || (KDE_VERSION_MINOR > 1)
  KMessageBox::sorry(0, i18n("no library specified!"));
#else
  KMsgBox::message(0, i18n("Sorry"), i18n("no library specified!"));
#endif
  return true;
}

int int9(double a)
{
  int i = (int) a;
  if (a < 0.0)
    i--;
  return i;
}

int minMaxFile(double *xmin, double *xmax, double *x, int n)
{
  *xmin = 1.0e300;
  *xmax = -*xmin;
  for (int i = 0; i < n; i++) {
    *xmin = QMIN(*xmin, x[i]);
    *xmax = QMAX(*xmax, x[i]);
  }
  return (*xmax == *xmin);
}

void expo(double a, int *ie, double *f)
{
  if((*ie = int9(log10(1.01 * a + 1.0e-37))) < 0)
    *ie = *ie - 1;
  *ie = -3 * (*ie / 3);
  *f = pow(10.0, *ie);
}

void expoFile(int *ie, double *f, double *x, int n)
{
  double xmin, xmax;

  minMaxFile(&xmin, &xmax, x, n);
  expo(QMAX(fabs(xmin), fabs(xmax)), ie, f);
}

void expoFunc(double (*fkt)(double, double *), int *ie, double *f,
              double xmin, double xmax, double dx, double *p)
{
  double ymax = 0.0;
  if (!dx)
    dx = 0.01 * (xmax - xmin);
  double x = xmin - dx;
  do {
    x += dx;
    if (fabs(x - xmax) < 0.9 * fabs(dx))
      x = xmax;
    ymax = QMAX(ymax, fabs(fkt(x, p)));
  } while (x != xmax);
  expo(ymax, ie, f);
}

bool autoSc(double *xmi, double *xma, double *tic, int *mtic, int *ndig,
            double xmin, double xmax, double fx, double c)
{
  double d, f;

  if (xmax == xmin)
    return true;
  d = c * fx * (xmax - xmin);
  f = pow(10.0, (- (float) (int9(log10(d + 1.0e-37)))));
  d = f * d;
  *mtic = 1;
  if (d > 1.414)
    *mtic = 2;
  if (d > 3.162)
    *mtic = 5;
  if (d > 7.071)
    *mtic = 10;
  *tic = *mtic / f;
  if ((*mtic == 1) || (*mtic == 10))
    *mtic = 2;
  if (!(*ndig = -int9(log10(*tic + 1.0e-37) + 1.0e-6)))
    *ndig = -1;
  *xmi = *tic * int9(fx * xmin / *tic + 1.0e-4);
  *xma = *tic * int9(fx * xmax / *tic + 0.9999);
  return false;
}

bool autoFile(double *xmi, double *xma, double *tic, int *mtic, int *ndig,
              int *ie, double *fn, double *x, int n, int norm, double c)
{
  double xmin, xmax;

  minMaxFile(&xmin, &xmax, x, n);
  if (norm)
    expo(QMAX(fabs(xmin), fabs(xmax)), ie, fn);
  else {
    *ie = 0;
    *fn = 1.0;
  }
  if (c)
    return autoSc(xmi, xma, tic, mtic, ndig, xmin, xmax, *fn, c);
  else {
    *xmi = xmin;
    *xma = xmax;
    return (xmin == xmax);
  }
}

bool autoFunc(double *ymi, double *yma, double *tic, int * mtic, int *ndig,
              double (*fkt)(double, double *), double xmin, double xmax,
              double dx, double *p, double fn, double c)
{
  double x, y, ymin, ymax;

  ymin = 1.0e300;
  ymax = -ymin;
  if(!dx)
    dx = 0.01 * (xmax - xmin);
  x = xmin - dx;
  do {
    x += dx;
    if (fabs(x - xmax) < 0.9 * fabs(dx))
      x = xmax;
    y = fn * fkt(x, p);
    ymin = QMIN(ymin, y);
    ymax = QMAX(ymax, y);
  } while (x != xmax);
  if (c)
    return autoSc(ymi, yma, tic, mtic, ndig, ymin, ymax, 1.0, c);
  else {
    *ymi = ymin;
    *yma = ymax;
    return (ymin == ymax);
  }
}
